# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.   Originates from datalad package distributed
#   under MIT license
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
from __future__ import annotations

from functools import wraps
import logging
import os
from os.path import abspath, exists, expanduser, expandvars, isabs
from os.path import join as opj
import platform
import shutil
import stat
import sys
import time
from types import TracebackType
from typing import Any

#
# Some useful variables
#
platform_system = platform.system()
on_windows = platform_system == "Windows"
on_osx = platform_system == "Darwin"
on_linux = platform_system == "Linux"


lgr = logging.getLogger("duecredit.utils")

#
# Little helpers
#


def is_interactive() -> bool:
    """Return True if all in/outs are tty"""
    if any(
        not hasattr(inout, "isatty") for inout in (sys.stdin, sys.stdout, sys.stderr)
    ):
        lgr.warning("Assuming non interactive session since isatty found missing")
        return False
    # TODO: check on windows if hasattr check would work correctly and add value:
    #
    return sys.stdin.isatty() and sys.stdout.isatty() and sys.stderr.isatty()


def expandpath(path: str, force_absolute: bool = True) -> str:
    """Expand all variables and user handles in a path.

    By default return an absolute path
    """
    path = expandvars(expanduser(path))
    if force_absolute:
        path = abspath(path)
    return path


def is_explicit_path(path: str) -> bool:
    """Return whether a path explicitly points to a location

    Any absolute path, or relative path starting with either '../' or
    './' is assumed to indicate a location on the filesystem. Any other
    path format is not considered explicit."""
    path = expandpath(path, force_absolute=False)
    return (
        isabs(path)
        or path.startswith(os.curdir + os.sep)
        or path.startswith(os.pardir + os.sep)
    )


def rotree(path: str, ro: bool = True, chmod_files: bool = True) -> None:
    """To make tree read-only or writable

    Parameters
    ----------
    path : string
      Path to the tree/directory to chmod
    ro : bool, optional
      Either to make it R/O (default) or RW
    chmod_files : bool, optional
      Either to operate also on files (not just directories)
    """
    if ro:

        def chmod(f: str) -> None:
            os.chmod(f, os.stat(f).st_mode & ~stat.S_IWRITE)

    else:

        def chmod(f: str) -> None:
            os.chmod(f, os.stat(f).st_mode | stat.S_IWRITE | stat.S_IREAD)

    for root, _, files in os.walk(path, followlinks=False):
        if chmod_files:
            for f in files:
                fullf = opj(root, f)
                # might be the "broken" symlink which would fail to stat etc
                if exists(fullf):
                    chmod(fullf)
        chmod(root)


def rmtree(
    path: str, chmod_files: str | bool = "auto", *args: Any, **kwargs: Any
) -> None:
    """To remove git-annex .git it is needed to make all files and directories writable again first

    Parameters
    ----------
    chmod_files : string or bool, optional
       Either to make files writable also before removal.  Usually it is just
       a matter of directories to have write permissions.
       If 'auto' it would chmod files on windows by default
    `*args` :
    `**kwargs` :
       Passed into shutil.rmtree call
    """
    # Give W permissions back only to directories, no need to bother with files
    if chmod_files == "auto":
        chmod_files_ = on_windows
    else:
        assert type(chmod_files) is bool
        chmod_files_ = chmod_files

    if not os.path.islink(path):
        rotree(path, ro=False, chmod_files=chmod_files_)
        shutil.rmtree(path, *args, **kwargs)
    else:
        # just remove the symlink
        os.unlink(path)


def rmtemp(f: str, *args: Any, **kwargs: Any) -> None:
    """Wrapper to centralize removing of temp files so we could keep them around

    It will not remove the temporary file/directory if DATALAD_TESTS_KEEPTEMP
    environment variable is defined
    """
    if not os.environ.get("DATALAD_TESTS_KEEPTEMP"):
        if not os.path.lexists(f):
            lgr.debug("Path %s does not exist, so can't be removed" % f)
            return
        lgr.log(5, "Removing temp file: %s" % f)
        # Can also be a directory
        if os.path.isdir(f):
            rmtree(f, *args, **kwargs)
        else:
            for i in range(10):
                try:
                    os.unlink(f)
                except OSError:
                    if i < 9:
                        time.sleep(0.1)
                        continue
                    else:
                        raise
                break
    else:
        lgr.info("Keeping temp file: %s" % f)


#
# Decorators
#


# Borrowed from pandas
# Copyright: 2011-2014, Lambda Foundry, Inc. and PyData Development Team
# License: BSD-3
def optional_args(decorator):
    """allows a decorator to take optional positional and keyword arguments.
    Assumes that taking a single, callable, positional argument means that
    it is decorating a function, i.e. something like this::

        @my_decorator
        def function(): pass

    Calls decorator with decorator(f, *args, **kwargs)"""

    @wraps(decorator)
    def wrapper(*args, **kwargs):
        def dec(f):
            return decorator(f, *args, **kwargs)

        is_decorating = not kwargs and len(args) == 1 and callable(args[0])
        if is_decorating:
            f = args[0]
            args = []
            return dec(f)
        else:
            return dec

    return wrapper


def never_fail(f):
    """Assure that function never fails -- all exceptions are caught"""

    @wraps(f)
    def wrapped_func(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except Exception as e:
            lgr.warning(
                "DueCredit internal failure while running %s: %r. "
                "Please report to developers at https://github.com/duecredit/duecredit/issues"
                % (f, e)
            )

    if os.environ.get("DUECREDIT_ALLOW_FAIL", False):
        return f
    else:
        return wrapped_func


def borrowdoc(cls, methodname: str | None = None, replace: str | None = None):
    """Return a decorator to borrow docstring from another `cls`.`methodname`

    Common use is to borrow a docstring from the class's method for an
    adapter function (e.g. sphere_searchlight borrows from Searchlight)

    Examples
    --------
    To borrow `__repr__` docstring from parent class `Mapper`, do::

       @borrowdoc(Mapper)
       def __repr__(self):
           ...

    Parameters
    ----------
    cls
      Usually a parent class
    methodname : None or str
      Name of the method from which to borrow.  If None, would use
      the same name as of the decorated method
    replace : None or str, optional
      If not None, then not entire docstring gets replaced but only the
      matching to "replace" value string
    """

    def _borrowdoc(method):
        """Decorator which assigns to the `method` docstring from another"""
        if methodname is None:
            other_method = getattr(cls, method.__name__)
        else:
            other_method = getattr(cls, methodname)
        if hasattr(other_method, "__doc__"):
            if not replace:
                method.__doc__ = other_method.__doc__
            else:
                method.__doc__ = method.__doc__.replace(replace, other_method.__doc__)
        return method

    return _borrowdoc


# TODO: just provide decorators for tempfile.mk* functions. This is ugly!
def get_tempfile_kwargs(
    tkwargs: dict[str, str] | None = None, prefix: str = "", wrapped=None
) -> dict[str, str]:
    """Updates kwargs to be passed to tempfile. calls depending on env vars"""
    # operate on a copy of tkwargs to avoid any side-effects
    if tkwargs is None:
        tkwargs = {}
    tkwargs_ = tkwargs.copy()

    # TODO: don't remember why I had this one originally
    # if len(targs)<2 and \
    if "prefix" not in tkwargs_:
        tkwargs_["prefix"] = "_".join(
            ["duecredit_temp"]
            + ([prefix] if prefix else [])
            + ([""] if (on_windows or not wrapped) else [wrapped.__name__])
        )

    directory = os.environ.get("DUECREDIT_TESTS_TEMPDIR")
    if directory and "dir" not in tkwargs_:
        tkwargs_["dir"] = directory

    return tkwargs_


#
# Context Managers
#

#
# Additional handlers
#
_sys_excepthook = sys.excepthook  # Just in case we ever need original one


def setup_exceptionhook() -> None:
    """Overloads default sys.excepthook with our exceptionhook handler.

    If interactive, our exceptionhook handler will invoke
    pdb.post_mortem; if not interactive, then invokes default handler.
    """

    def _duecredit_pdb_excepthook(
        exc_type: type[BaseException],
        exc_value: BaseException,
        exc_tb: TracebackType | None = None,
    ) -> None:
        if is_interactive():
            import pdb
            import traceback

            traceback.print_exception(exc_type, exc_value, exc_tb)
            print
            pdb.post_mortem(exc_tb)
        else:
            lgr.warn("We cannot setup exception hook since not in interactive mode")
            # we are in interactive mode or we don't have a tty-like
            # device, so we call the default hook
            # sys.__excepthook__(exc_type, exc_value, exc_tb)
            _sys_excepthook(exc_type, exc_value, exc_tb)

    sys.excepthook = _duecredit_pdb_excepthook



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Module/app to automate collection of relevant to analysis publications.

Please see README.md shipped along with duecredit to get a better idea about
its functionality
"""

from .dueswitch import due
from .entries import BibTeX, Doi, Text, Url
from .version import __release_date__, __version__

__all__ = ["Doi", "BibTeX", "Url", "Text", "due", "__version__", "__release_date__"]



#!/usr/bin/env python
"""
duecredit -- publications (donations, etc) tracer
"""

from datetime import datetime
import os.path
import re
from subprocess import PIPE, Popen
import sys

from setuptools import find_packages, setup

# Adopted from citeproc-py
#  License: BSD-2
#  Copyright 2011-2013 Brecht Machiels

PACKAGE = "duecredit"
VERSION_FILE = PACKAGE + "/version.py"


def make_pep440_compliant(version: str, local_prefix: str) -> str:
    """Convert the version into a PEP440 compliant version."""
    public_version_re = re.compile(
        r"^([0-9][0-9.]*(?:(?:a|b|rc|.post|.dev)[0-9]+)*)\+?"
    )
    _, public, local = public_version_re.split(version, maxsplit=1)
    if not local:
        return version
    sanitized_local = re.sub("[+~-]+", ".", local).strip(".")
    pep440_version = f"{public}+{local_prefix}{sanitized_local}"
    assert re.match(
        "^[a-zA-Z0-9.]+$", sanitized_local
    ), f"'{pep440_version}' not PEP440 compliant"
    return pep440_version


# retrieve the version number from git or VERSION_FILE
# inspired by http://dcreager.net/2010/02/10/setuptools-git-version-numbers/

try:
    if os.path.exists("debian/copyright"):
        print("Generating version.py out of debian/copyright information")
        # building debian package. Deduce version from debian/copyright
        with open("debian/changelog") as f:
            lines = f.readlines()
        __version__ = make_pep440_compliant(lines[0].split()[1].strip("()"), "debian.")
        # TODO: unify format whenever really bored ;)
        __release_date__ = re.sub(
            r"^ -- .*>\s*(.*)",
            r"\1",
            list(filter(lambda x: x.startswith(" -- "), lines))[0].rstrip(),
        )
    else:
        print("Attempting to get version number from git...")
        git = Popen(
            ["git", "describe", "--abbrev=4", "--dirty"], stdout=PIPE, stderr=sys.stderr
        )
        if git.wait() != 0:
            raise OSError
        assert git.stdout
        line = git.stdout.readlines()[0].strip().decode("ascii")
        if line.count("-") >= 2:
            # we should parse it to make version compatible with PEP440
            # unfortunately we wouldn't be able to include git treeish
            # into the version, and thus can have collisions. So let's
            # release from master only
            line = ".dev".join(line.split("-")[:2])
        __version__ = line
        __release_date__ = datetime.now().strftime("%b %d %Y, %H:%M:%S")
    with open(VERSION_FILE, "w") as version_file:
        version_file.write(f"__version__ = '{__version__}'\n")
        version_file.write(f"__release_date__ = '{__release_date__}'\n")
except OSError:
    print("Assume we are running from a source distribution.")
    # read version from VERSION_FILE
    if os.path.exists(VERSION_FILE):
        with open(VERSION_FILE) as version_file:
            code = compile(version_file.read(), VERSION_FILE, "exec")
            exec(code, locals(), globals())
    else:
        __version__ = "0.0.0.dev"
print("Version: %s" % __version__)

with open("README.md", encoding="utf-8") as f:
    README = f.read()

setup(
    name=PACKAGE,
    version=__version__,
    packages=find_packages(),
    python_requires=">=3.8",
    install_requires=[
        "citeproc-py>=0.4",
        "looseversion",
        "packaging",
        "requests",
    ],
    extras_require={"tests": ["pytest", "pytest-cov", "vcrpy", "contextlib2"]},
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "duecredit=duecredit.cmdline.main:main",
        ],
    },
    author="Yaroslav Halchenko, Matteo Visconti di Oleggio Castello",
    author_email="yoh@onerussian.com",
    description="Publications (and donations) tracer",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/duecredit/duecredit",
    keywords=["citation tracing"],
    license="2-clause BSD License",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Environment :: Other Environment",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Legal Industry",
        "Intended Audience :: Other Audience",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Documentation",
        "Topic :: Software Development :: Documentation",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
from __future__ import annotations

"""Module to help maintain a registry of versions for external modules etc
"""
from collections.abc import KeysView
from importlib.metadata import version as metadata_version
from os import linesep
import sys
from types import ModuleType
from typing import Any

from looseversion import LooseVersion
from packaging.version import Version


# To depict an unknown version, which can't be compared by mistake etc
class UnknownVersion:
    """For internal use"""

    def __str__(self) -> str:
        return "UNKNOWN"

    def __eq__(self, other: Any) -> bool:
        if other is self:
            return True
        raise TypeError("UNKNOWN version is not comparable")


class ExternalVersions:
    """Helper to figure out/use versions of the external modules.

    It maintains a dictionary of `packaging.version.Version`s to make
    comparisons easy.  If a version string doesn't conform to Version,
    LooseVersion will be used.  If a version can't be deduced for a module,
    'None' is assigned
    """

    UNKNOWN = UnknownVersion()

    def __init__(self) -> None:
        self._versions: dict[str, Version | LooseVersion | UnknownVersion] = {}

    @classmethod
    def _deduce_version(
        klass, module: ModuleType
    ) -> Version | LooseVersion | UnknownVersion:
        version = None
        for attr in ("__version__", "version"):
            if hasattr(module, attr):
                version = getattr(module, attr)
                break

        if isinstance(version, tuple) or isinstance(version, list):
            #  Generate string representation
            version = ".".join(str(x) for x in version)

        if not version:
            # Try importlib.metadata
            # module name might be different, and I found no way to
            # deduce it for citeproc which comes from "citeproc-py"
            # distribution
            modname = module.__name__
            try:
                version = metadata_version(
                    {"citeproc": "citeproc-py"}.get(modname, modname)
                )
            except Exception:
                pass  # oh well - no luck either

        if version:
            try:
                return Version(version)
            except ValueError:
                # let's then go with Loose one
                return LooseVersion(version)
        else:
            return klass.UNKNOWN

    def __getitem__(
        self, module: Any
    ) -> Version | LooseVersion | UnknownVersion | None:
        # when ran straight in its source code -- fails to discover nipy's version.. TODO
        # if module == 'nipy':
        if not isinstance(module, str):
            modname = module.__name__
        else:
            modname = module
            module = None

        if modname not in self._versions:
            if module is None:
                if modname not in sys.modules:
                    try:
                        module = __import__(modname)
                    except ImportError:
                        return None
                else:
                    module = sys.modules[modname]

            self._versions[modname] = self._deduce_version(module)

        return self._versions.get(modname, self.UNKNOWN)

    def keys(self) -> KeysView[str]:
        """Return names of the known modules"""
        return self._versions.keys()

    def __contains__(self, item: str) -> bool:
        return item in self._versions

    @property
    def versions(self) -> dict[str, Version | LooseVersion | UnknownVersion]:
        """Return dictionary (copy) of versions"""
        return self._versions.copy()

    def dumps(self, indent: bool | str = False, preamble: str = "Versions:") -> str:
        """Return listing of versions as a string

        Parameters
        ----------
        indent: bool or str, optional
          If set would instruct on how to indent entries (if just True, ' '
          is used). Otherwise returned in a single line
        preamble: str, optional
          What preamble to the listing to use
        """
        items = ["{}={}".format(k, self._versions[k]) for k in sorted(self._versions)]
        out = "%s" % preamble
        if indent:
            indent_ = " " if indent is True else indent
            out += (linesep + indent_).join([""] + items) + linesep
        else:
            out += " " + " ".join(items)
        return out


external_versions = ExternalVersions()



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
from __future__ import annotations

from io import StringIO
import sys

import pytest
from pytest import MonkeyPatch

from .. import __version__
from ..cmdline import main


def test_import() -> None:
    import duecredit.cmdline  # noqa: F401
    import duecredit.cmdline.main  # noqa: F401


def test_main_help(monkeypatch: MonkeyPatch) -> None:
    # Patch stdout
    fakestdout = StringIO()
    monkeypatch.setattr(sys, "stdout", fakestdout)

    pytest.raises(SystemExit, main.main, ["--help"])
    assert fakestdout.getvalue().lstrip().startswith("Usage: ")


def test_main_version(monkeypatch: MonkeyPatch) -> None:
    # Patch stdout or stderr for different Python versions -- catching both
    fakestdout = StringIO()
    fakeout = "stdout"
    monkeypatch.setattr(sys, fakeout, fakestdout)

    pytest.raises(SystemExit, main.main, ["--version"])
    assert (fakestdout.getvalue()).split("\n")[0] == "duecredit %s" % __version__


# smoke test the cmd_summary
# TODO: carry sample .duecredit.p, point to that file, monkeypatch TextOutput and BibTeXOutput .dumps
def test_smoke_cmd_summary() -> None:
    main.main(["summary"])


def test_cmd_test() -> None:  # test the not implemented cmd_test
    pytest.raises(SystemExit, main.main, ["test"])



"""Some test submodule"""
from typing import Any


def testfunc(arg1: Any, kwarg1: Any = None) -> str:
    """testfunc docstring"""
    return "testfunc: {}, {}".format(arg1, kwarg1)


class TestClass:
    def testmeth(self, arg1: Any, kwarg1: Any = None) -> str:
        return "testmeth: {}, {}".format(arg1, kwarg1)
